from __future__ import annotations

from pathlib import Path
from typing import Any, Callable, Collection

import hypothesis.strategies as st
import pytest
from attr import asdict
from hypothesis import HealthCheck, given, settings

from negmas import (
    AspirationNegotiator,
    Issue,
    MappingUtilityFunction,
    MechanismState,
    Negotiator,
    NegotiatorMechanismInterface,
    RenegotiationRequest,
    SAOMechanism,
)
from negmas.events import Event, EventSink, EventSource
from negmas.helpers import unique_name
from negmas.outcomes import make_issue
from negmas.situated import Action, Agent, Breach, Contract, World

results = []  # will keep results not to use printing


class DummyWorld(World):
    def __init__(self, n_steps=10000, negotiation_speed=20, **kwargs):
        super().__init__(
            n_steps=n_steps,
            negotiation_speed=negotiation_speed,
            neg_n_steps=10,
            neg_time_limit=10,
            **kwargs,
        )
        self.the_agents = []

    def complete_contract_execution(
        self, contract: Contract, breaches: list[Breach], resolved: bool
    ) -> None:
        pass

    def get_contract_finalization_time(self, contract: Contract) -> int:
        return self.current_step + 1

    def get_contract_execution_time(self, contract: Contract) -> int:
        return self.current_step

    def contract_size(self, contract: Contract) -> float:
        return 0.0

    def join(self, x: Agent, simulation_priority: int = 0):
        super().join(x=x, simulation_priority=simulation_priority)
        self.the_agents.append(x)

    def delete_executed_contracts(self) -> None:
        pass

    def executable_contracts(self) -> Collection[Contract]:
        return []

    def post_step_stats(self):
        pass

    def pre_step_stats(self):
        pass

    def order_contracts_for_execution(
        self, contracts: Collection[Contract]
    ) -> Collection[Contract]:
        return contracts

    def contract_record(self, contract: Contract) -> dict[str, Any]:
        return asdict(contract)

    def start_contract_execution(self, contract: Contract) -> set[Breach]:
        return set()

    def _process_breach(
        self, contract: Contract, breaches: list[Breach], force_immediate_signing=True
    ) -> Contract | None:
        return None

    def breach_record(self, breach: Breach):
        return breach.__dict__

    def execute_action(
        self, action: Action, agent: Agent, callback: Callable = None
    ) -> bool:
        return True

    def get_private_state(self, agent: Agent) -> Any:
        s = {"partners": [_ for _ in self.the_agents if _ is not agent]}
        return s

    def simulation_step(self, stage: int):
        pass


class DummyAgent(Agent):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.id = self.name
        self.__current_step = 0

    def init(self):
        pass

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        issues: list[Issue],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        role: str | None,
        req_id: str | None,
    ) -> Negotiator | None:
        negotiator = AspirationNegotiator(
            preferences=MappingUtilityFunction(
                mapping=lambda x: 1.0 - x[0] / 10.0, issues=issues
            )
        )
        return negotiator

    def on_neg_request_rejected(self, req_id: str, by: list[str] | None):
        pass

    def on_neg_request_accepted(
        self, req_id: str, mechanism: NegotiatorMechanismInterface
    ):
        pass

    def on_negotiation_failure(
        self,
        partners: list[str],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        state: MechanismState,
    ) -> None:
        pass

    def on_negotiation_success(
        self, contract: Contract, mechanism: NegotiatorMechanismInterface
    ) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: list[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> str | None:
        return self.id

    def on_contract_executed(self, contract: Contract) -> None:
        pass

    def on_contract_breached(
        self, contract: Contract, breaches: list[Breach], resolution: Contract | None
    ) -> None:
        pass

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: list[Breach]
    ) -> RenegotiationRequest | None:
        return None

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: list[Breach], agenda: RenegotiationRequest
    ) -> Negotiator | None:
        return None

    def step(self):
        global results
        if self.awi is None:
            return
        self.__current_step = self.awi.current_step
        if (self.__current_step == 2 and self.name.endswith("1")) or (
            self.__current_step == 4 and self.name.endswith("2")
        ):
            issues = [make_issue(10, name="i1")]
            partners = self.awi.state["partners"]
            self._request_negotiation(
                partners=[_.name for _ in partners] + [self.name], issues=issues
            )
            results.append(f"{self.name} started negotiation with {partners[0].name}")
        results.append(f"{self.name}: step {self.__current_step}")


class ExceptionAgent(Agent):
    def __init__(self, name=None):
        super().__init__(name=name)
        self.id = name
        self.__current_step = 0

    def init(self):
        pass

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        issues: list[Issue],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        role: str | None,
        req_id: str | None,
    ) -> Negotiator | None:
        negotiator = AspirationNegotiator(
            preferences=MappingUtilityFunction(
                mapping=lambda x: 1.0 - x[0] / 10.0,
                issues=issues,
            ),
        )
        return negotiator

    def on_neg_request_rejected(self, req_id: str, by: list[str] | None):
        pass

    def on_neg_request_accepted(
        self, req_id: str, mechanism: NegotiatorMechanismInterface
    ):
        pass

    def on_negotiation_failure(
        self,
        partners: list[str],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        state: MechanismState,
    ) -> None:
        pass

    def on_negotiation_success(
        self, contract: Contract, mechanism: NegotiatorMechanismInterface
    ) -> None:
        pass

    def on_contract_signed(self, contract: Contract) -> None:
        pass

    def on_contract_cancelled(self, contract: Contract, rejectors: list[str]) -> None:
        pass

    def sign_contract(self, contract: Contract) -> str | None:
        return self.id

    def on_contract_executed(self, contract: Contract) -> None:
        pass

    def on_contract_breached(
        self, contract: Contract, breaches: list[Breach], resolution: Contract | None
    ) -> None:
        pass

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: list[Breach]
    ) -> RenegotiationRequest | None:
        return None

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: list[Breach], agenda: RenegotiationRequest
    ) -> Negotiator | None:
        return None

    def step(self):
        global results
        if self.awi is None:
            return
        self.__current_step = self.awi.current_step
        if (self.__current_step == 2 and self.name.endswith("1")) or (
            self.__current_step == 4 and self.name.endswith("2")
        ):
            issues = [make_issue(10, name="i1")]
            partners = self.awi.state["partners"]
            self._request_negotiation(
                partners=[_.name for _ in partners] + [self.name], issues=issues
            )
            results.append(f"{self.name} started negotiation with {partners[0].name}")
        raise ValueError("error")
        results.append(f"{self.name}: step {self.__current_step}")


def test_world_has_times():
    import time

    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    _strt = time.perf_counter()
    world.run()
    t = time.perf_counter() - _strt
    for aid in world.agents.keys():
        assert 0.0 < world.times[aid] < t
    assert 0.0 < sum(world.times.values()) < t


def test_world_records_exceptions():
    pass

    world = DummyWorld(n_steps=10, ignore_agent_exceptions=True)
    world.join(ExceptionAgent("A1"))
    world.join(ExceptionAgent("A2"))
    world.run()
    assert len(world.simulation_exceptions) == 0
    assert len(world.contract_exceptions) == 0
    assert len(world.mechanism_exceptions) == 0
    for aid in world.agents.keys():
        assert len(world.negotiator_exceptions[aid]) == 0.0
        assert len(world.agent_exceptions[aid]) == 10
        assert sum(world.n_negotiator_exceptions.values()) == 0
        assert sum(world.n_agent_exceptions.values()) == 20
        a = list(world.n_agent_exceptions.values())
        assert a[0] == a[1] == 10
        assert sum(world.n_total_simulation_exceptions.values()) == 0
        assert sum(world.n_total_agent_exceptions.values()) == 20


def test_world_runs_with_some_negs():
    global results
    results = []
    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    world.run()
    assert "A1: step 1" in results, "first step logged"
    assert "A2: step 1" in results, "first step logged"
    assert "A1 started negotiation with A2" in results, "negotiation started"
    assert f"A1: step {world.n_steps-1}" in results, "last step logged"
    assert f"A2: step {world.n_steps-1}" in results, "last step logged"
    assert len(world.saved_contracts) == 2
    assert sum(world.stats["n_negotiations"]) == 2


def test_config_reader_with_a_world():

    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(scope=globals(), config={"n_steps": 10})
    assert world.bulletin_board is not None
    assert world.n_steps == 10


def test_config_reader_with_a_world_with_enum():
    world = DummyWorld()
    assert world.bulletin_board is not None
    assert world.n_steps == 10000

    world = DummyWorld.from_config(
        scope=globals(), config={"n_steps": 10, "negotiation_speed": 2}
    )
    assert world.bulletin_board is not None
    assert world.n_steps == 10
    assert world.negotiation_speed == 2


def test_world_picklable(tmp_path):
    import dill as pickle

    world = DummyWorld()
    world.step()
    world.step()
    file = tmp_path / "world.pckl"

    with open(file, "wb") as f:
        pickle.dump(world, f)
    with open(file, "rb") as f:
        w = pickle.load(f)
    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.run()


@settings(deadline=None, suppress_health_check=[HealthCheck.function_scoped_fixture])
@given(
    single_checkpoint=st.booleans(),
    checkpoint_every=st.integers(0, 6),
    exist_ok=st.booleans(),
)
def test_world_auto_checkpoint(tmp_path, single_checkpoint, checkpoint_every, exist_ok):
    import shutil

    new_folder: Path = tmp_path / unique_name("empty", sep="")
    new_folder.mkdir(parents=True, exist_ok=True)
    shutil.rmtree(new_folder)
    new_folder.mkdir(parents=True, exist_ok=True)
    filename = "mechanism"
    n_steps = 20

    world = DummyWorld(
        n_steps=n_steps,
        checkpoint_every=checkpoint_every,
        checkpoint_folder=new_folder,
        checkpoint_filename=filename,
        extra_checkpoint_info=None,
        exist_ok=exist_ok,
        single_checkpoint=single_checkpoint,
    )

    world.run()

    if 0 < checkpoint_every <= n_steps:
        if single_checkpoint:
            assert len(list(new_folder.glob("*"))) == 2, print(
                f"World ran for: {world.current_step}"
            )
        else:
            assert len(list(new_folder.glob("*"))) >= 2 * (
                max(1, world.current_step // checkpoint_every)
            )
    elif checkpoint_every > n_steps:
        assert len(list(new_folder.glob("*"))) == 2
    else:
        assert len(list(new_folder.glob("*"))) == 0


def test_world_checkpoint(tmp_path):
    world = DummyWorld()
    world.step()
    world.step()

    file_name = world.checkpoint(tmp_path)

    info = DummyWorld.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 2
    assert info["type"].endswith("DummyWorld")
    assert info["id"] == world.id
    assert info["name"] == world.name

    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    world = w
    file_name = world.checkpoint(tmp_path)
    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.step()

    world = w
    file_name = world.checkpoint(tmp_path)
    w = DummyWorld.from_checkpoint(file_name)

    assert world.current_step == w.current_step
    assert world.agents == w.agents
    assert w.bulletin_board is not None
    assert w.n_steps == world.n_steps
    assert w.negotiation_speed == world.negotiation_speed

    w.run()


def test_agent_checkpoint(tmp_path):

    a = DummyAgent(name="abcd")

    a.step_()

    file_name = a.checkpoint(tmp_path, info={"a": 3})

    info = SAOMechanism.checkpoint_info(file_name)
    assert isinstance(info["time"], str)
    assert info["step"] == 1
    assert info["type"].endswith("DummyAgent")
    assert info["id"] == a.id
    assert info["name"] == a.name == "abcd"
    assert info["a"] == 3

    b = a.from_checkpoint(file_name)

    assert a.id == b.id

    b.step_()


def test_agent_checkpoint_in_world(tmp_path):
    world = DummyWorld(n_steps=10)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    assert len(world.agents) == 2
    for a in world.agents.values():
        a.step_()
        file_name = a.checkpoint(tmp_path, info={"a": 3})

        info = SAOMechanism.checkpoint_info(file_name)
        assert isinstance(info["time"], str)
        assert info["step"] == 1
        assert info["type"].endswith("DummyAgent")
        assert info["id"] == a.id
        assert info["name"] == a.name
        assert info["a"] == 3

        b = a.from_checkpoint(file_name)

        assert a.id == b.id

        b.step_()


class MyMonitor(EventSink):
    def on_event(self, event: Event, sender: EventSource):
        assert event.type == "agent-joined"


def test_cannot_start_a_neg_with_no_outcomes():
    world = DummyWorld(n_steps=10)
    a, b = DummyAgent(name="a"), DummyAgent(name="b")
    world.join(a)
    world.join(b)
    with pytest.raises(ValueError):
        a.awi.request_negotiation_about(
            issues=[make_issue((1, 0))], partners=[a.id, b.id], req_id="1234"
        )


def test_world_monitor():
    monitor = MyMonitor()
    world = DummyWorld(n_steps=10)
    world.register_listener("agent-joined", listener=monitor)
    world.join(DummyAgent("A1"))
    world.join(DummyAgent("A2"))
    assert len(world.agents) == 2
    world.run()


if __name__ == "__main__":
    pytest.main(args=[__file__])



"""
This module defines the interfaces to all negotiation agents (negotiators)
in negmas.
"""
from .common import *
from .negotiator import *
from .controller import *
from .controlled import *
from .components import *
from .helpers import *
from .simple import *
from .modular import *

__all__ = (
    common.__all__
    + negotiator.__all__
    + controller.__all__
    + controlled.__all__
    + components.__all__
    + simple.__all__
    + helpers.__all__
    + modular.__all__
)



from __future__ import annotations

import os
import shutil
from os import walk
from pathlib import Path

import pkg_resources
import pytest
from pytest import mark

from negmas import load_genius_domain_from_folder
from negmas.genius import genius_bridge_is_running
from negmas.inout import Scenario
from negmas.outcomes import enumerate_issues
from negmas.preferences.crisp.nonlinear import HyperRectangleUtilityFunction
from negmas.preferences.discounted import DiscountedUtilityFunction
from negmas.sao import AspirationNegotiator


@pytest.fixture
def scenarios_folder():
    return pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/scenarios"
    )


# todo: get these to work
SCENARIOS_TO_IGNORE = [
    "S-1NIKFRT-3",
    "IntegerDomain",
    "S-1NIKFRT-2",
    "S-1NAGUNL-128",
    "S-1NAGUNL-255",
    "50issueswithRV",
    "50issues",
    "30issuesDiscounted",
    "30issueswithRV",
    "30issues",
    "50issuesDiscounted",
    "30issuesDiscountedwithRV",
    "50issuesDiscountedwithRV",
    "AgentHp2",
    "web_service",
    "four_issues",
    "AMPOvsCity",
    "laptopdomain",
    "S-1NIKFRT-1",
    "S-1NAGUNL-114",
    "laptopdomainNoBayes",
    "inheritancedomain",
    "10issues",
    "10issuesDiscounted",
    "10issueswithRV",
    "10issuesDiscountedwithRV",
    "group9-vacation",
    "group9-killer_robot",
    "FitnessA",
    "FitnessB",
    "FitnessC",
    "Grandma",
    "YXAgent",
    # "group5-car_domain",
    # "group2-new_sporthal",
    # "group12-symposium",
    # "group11-car_purchase",
    # "group9-vacation",
    # "group8-holiday",
]


def get_all_scenarios():
    base = Path(__file__).parent / "data" / "scenarios"
    data = []
    for root, dirs, files in walk(base):
        if len(files) == 0 or len(dirs) != 0:
            continue
        if root.split("/")[-1] in SCENARIOS_TO_IGNORE:
            continue
        data.append(root)
    return data


def test_reading_writing_linear_preferences(tmp_path):
    from negmas.preferences import LinearAdditiveUtilityFunction, UtilityFunction

    base_folder = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    domain = load_genius_domain_from_folder(
        base_folder,
    )
    ufuns, issues = domain.ufuns, domain.issues
    for ufun in ufuns:
        assert isinstance(ufun, LinearAdditiveUtilityFunction)
        dst = tmp_path / "tmp.xml"
        UtilityFunction.to_genius(ufun, issues=issues, file_name=dst)
        print(str(dst))
        ufun2, _ = UtilityFunction.from_genius(dst, issues=issues)
        try:
            os.unlink(dst)
        except:
            pass
        assert isinstance(ufun2, LinearAdditiveUtilityFunction)
        for outcome in enumerate_issues(issues):
            assert abs(ufun2(outcome) - ufun(outcome)) < 1e-3


def test_importing_file_without_exceptions(scenarios_folder):
    folder_name = scenarios_folder + "/other/S-1NIKFRT-1"
    load_genius_domain_from_folder(folder_name)


def test_simple_run_with_aspiration_agents():
    file_name = pkg_resources.resource_filename(
        "negmas", resource_name="tests/data/Laptop"
    )
    assert os.path.exists(file_name)
    domain = Scenario.from_genius_folder(Path(file_name))
    assert domain
    domain.to_single_issue()
    mechanism = domain.make_session(AspirationNegotiator, n_steps=100, time_limit=30)
    assert mechanism is not None
    mechanism.run()


def compared_two_domains(domain, domain2):
    from negmas.genius import AgentX, Atlas3

    assert len(domain.issues) == len(domain2.issues)
    for i1, i2 in zip(domain.issues, domain2.issues):
        assert (
            i1.cardinality == i2.cardinality
            and i1.type == i2.type
            and i1.value_type == i2.value_type
        )

    assert len(domain.ufuns) == len(domain2.ufuns)

    for u1, u2 in zip(domain.ufuns, domain2.ufuns):
        while isinstance(u1, DiscountedUtilityFunction):
            u1 = u1.ufun
        while isinstance(u2, DiscountedUtilityFunction):
            u2 = u2.ufun
        if isinstance(u1, HyperRectangleUtilityFunction) or isinstance(
            u2, HyperRectangleUtilityFunction
        ):
            continue
        dm = domain.agenda.to_discrete(5)
        for i, w in enumerate(dm):
            if i > 10_000:
                return
            u1_, u2_ = u1(w), u2(w)
            assert isinstance(u1_, float)
            assert isinstance(u2_, float)
            assert abs(u1_ - u2_) < 1e-3, f"{str(u1)}\n{str(u2)}"

    for ufun in domain.ufuns:
        if isinstance(ufun, HyperRectangleUtilityFunction):
            continue
        m = domain.make_session(n_steps=100, name=domain.agenda.name)
        assert m is not None
        if not genius_bridge_is_running():
            continue
        n1 = Atlas3(domain_file_name=m.name, preferences=ufun)
        n2 = AgentX(domain_file_name=m.name, utility_file_name=ufun.name)
        m.add(n1)
        m.add(n2)
        u1, u2 = n1.ufun, n2.ufun
        assert u1 and u2
        outcomes = m.discrete_outcomes(max_cardinality=100)
        for outcome in outcomes:
            u1_, u2_ = u1(outcome), u2(outcome)
            assert isinstance(u1_, float)
            assert isinstance(u2_, float)
            assert abs(u1_ - u2_) < 1e-3
        n1.destroy_java_counterpart()
        n2.destroy_java_counterpart()


def do_enc_dec_trial(tmp, folder_name, with_discounting=True):
    domain = Scenario.from_genius_folder(folder_name, safe_parsing=False)
    assert domain
    domain = domain.remove_discounting()
    print(f"{str(folder_name)}\n-> {str(tmp)}")
    domain.to_genius_folder(tmp)
    domain2 = Scenario.from_genius_folder(tmp)
    assert domain2
    if not with_discounting:
        domain2 = domain2.remove_discounting()
    compared_two_domains(domain, domain2)
    try:
        shutil.rmtree(tmp)
    except:
        pass


@mark.xfail(
    run=False, reason="Known to fail. It is the int/discrete issue ambiguity in Genius"
)
@pytest.mark.parametrize("disc", [True, False])
def test_encoding_decoding_example_AMPOvsCity(tmp_path, disc):
    folder_name = Path(__file__).parent / "data" / "scenarios" / "other" / "AMPOvsCity"
    do_enc_dec_trial(tmp_path / "tmp", folder_name, disc)


@pytest.mark.parametrize("disc", [True, False])
def test_encoding_decoding_example_group_8_holiday(tmp_path, disc):
    folder_name = (
        Path(__file__).parent
        / "data"
        / "scenarios"
        / "anac"
        / "y2015"
        / "group8-holiday"
    )
    do_enc_dec_trial(tmp_path / "tmp", folder_name, disc)


@pytest.mark.parametrize("folder_name", get_all_scenarios())
def test_encoding_decoding_all_without_discounting(tmp_path, folder_name):
    do_enc_dec_trial(tmp_path / "tmp", folder_name, False)


@pytest.mark.parametrize("folder_name", get_all_scenarios())
def test_encoding_decoding_all_with_discounting(tmp_path, folder_name):
    do_enc_dec_trial(tmp_path / "tmp", folder_name, True)


# @pytest.mark.parametrize("folder_name", get_all_scenarios())
# def test_importing_all_single_issue_without_exceptions(folder_name):
#     # def test_encoding_decoding_all_without_discounting():
#     # folder_name = "/Users/yasser/code/projects/negmas/tests/data/scenarios/anac/y2012/FitnessC"
#     d2 = Domain.from_genius_folder(folder_name, safe_parsing=False)
#     d2.discretize()
#     assert isinstance(d2.agenda, DiscreteCartesianOutcomeSpace)
#     n = d2.agenda.cardinality
#     if n < 10_000:
#         d2.to_single_issue()
#         assert d2.agenda.cardinality == n or d2.agenda.cardinality == float("inf")



#!/usr/bin/env python
"""A set of utilities that can be used by agents developed for the platform.

This set of utlities can be extended but must be backward compatible for at
least two versions
"""
from __future__ import annotations

import base64
import json
import os
import pathlib
from os import PathLike
from pathlib import Path
from typing import Any, Iterable

import dill as pickle
import inflect
import numpy as np
import pandas as pd
import stringcase
import yaml

from negmas import warnings
from negmas.config import NEGMAS_CONFIG

from .types import TYPE_START, get_class, get_full_type_name, is_jsonable

__all__ = [
    "is_nonzero_file",
    "ConfigReader",
    "DEFAULT_DUMP_EXTENSION",
    "dump",
    "load",
    "add_records",
    "TYPE_START",
]
# conveniently named classes
BYTES_START = "__BYTES__:"
PATH_START = "__PATH__:"
"""Maps from a single issue to a Negotiator function."""

DEFAULT_DUMP_EXTENSION = NEGMAS_CONFIG.get("default_dump_extension", "json")


def is_nonzero_file(fpath: PathLike) -> bool:
    """Whether or not the path is for an existing nonzero file.

    Args:
        fpath: path to the file to test. It accepts both str and pathlib.Path

    """
    return os.path.isfile(fpath) and os.path.getsize(fpath) > 0


_inflect_engine = inflect.engine()


class ConfigReader:
    @classmethod
    def _split_key(cls, key: str) -> tuple[str, str | None]:
        """Splits the key into a key name and a class name

        Remarks:

            - Note that if the given key has multiple colons the first two will be parsed as key name: class name and
              the rest will be ignored. This can be used to add comments

        """
        keys = key.split(":")
        if len(keys) == 1:
            return keys[0], None
        else:
            return keys[0], keys[1]

    @classmethod
    def _parse_children_config(cls, children, scope):
        """Parses children in the given scope"""
        remaining_children = {}
        myconfig = {}
        setters = []
        for key, v in children.items():
            k, class_name = cls._split_key(key)
            if isinstance(v, dict):
                if class_name is None:
                    class_name = stringcase.pascalcase(k)
                the_class = get_class(class_name=class_name, scope=scope)
                assert hasattr(
                    the_class, "from_config"
                ), f"{the_class} does not have a from_config attribute"
                obj, obj_children = the_class.from_config(  # type: ignore
                    config=v,
                    ignore_children=False,
                    try_parsing_children=True,
                    scope=scope,
                )
                if obj_children is not None and len(obj_children) > 0:
                    remaining_children[k] = obj_children
                setter_name = "set_" + k
                if hasattr(cls, setter_name):
                    setters.append((setter_name, obj))
                else:
                    myconfig[k] = obj
            elif isinstance(v, Iterable) and not isinstance(v, str):
                singular = _inflect_engine.singular_noun(k)
                if singular is False:
                    singular = k
                if class_name is None:
                    class_name = stringcase.pascalcase(singular)
                setter_name = "set_" + k
                objs = []
                for current in list(v):
                    the_class = get_class(class_name=class_name, scope=scope)
                    assert hasattr(
                        the_class, "from_config"
                    ), f"{the_class} does not have a from_config attribute"
                    obj = the_class.from_config(  # type: ignore
                        config=current,
                        ignore_children=True,
                        try_parsing_children=True,
                        scope=scope,
                    )
                    objs.append(obj)
                if hasattr(cls, setter_name):
                    setters.append((setter_name, objs))
                else:
                    myconfig[k] = objs
            else:
                # not a dictionary and not an iterable.
                remaining_children[k] = v

        return myconfig, remaining_children, setters

    @classmethod
    def read_config(
        cls, config: str | dict, section: str | None = None
    ) -> dict[str, Any]:
        """
        Reads the configuration from a file or a dict and prepares it for parsing

        Args:
            config: Either a file name or a dictionary
            section: A section in the file or a key in the dictionary to use for loading params

        Returns:

            A dict ready to be parsed by from_config

        Remarks:


        """
        if isinstance(config, str):
            # If config is a string, assume it is a file and read it from the appropriate location
            def exists(nm):
                return os.path.exists(nm) and not os.path.isdir(nm)

            if not exists(config):
                name = pathlib.Path("./") / pathlib.Path(config)
                if exists(name):
                    config = str(name.absolute())
                else:
                    name = (pathlib.Path("./.negmas") / config).absolute()
                    if exists(name):
                        config = str(name)
                    else:
                        name = (
                            pathlib.Path(os.path.expanduser("~/.negmas")) / config
                        ).absolute()
                        if exists(name):
                            config = str(name)
                        else:
                            raise ValueError(f"Cannot find config in {config}.")
            with open(config) as f:
                if config.endswith(".json"):
                    config = json.load(f)
                elif config.endswith(".cfg"):
                    config = eval(f.read())
                elif config.endswith(".yaml") or config.endswith(".yml"):
                    config = yaml.safe_load(f)
                else:
                    raise ValueError(f"Cannot parse {config}")

        if section is not None:
            config = config[section]  # type: ignore

        return config  # type: ignore

    @classmethod
    def from_config(
        cls,
        config: str | dict,
        section: str | None = None,
        ignore_children: bool = True,
        try_parsing_children: bool = True,
        scope=None,
    ):
        """
        Creates an object of this class given the configuration info

        Args:
            config: Either a file name or a dictionary
            section: A section in the file or a key in the dictionary to use for loading params
            ignore_children: If true then children will be ignored and there will be a single return
            try_parsing_children: If true the children will first be parsed as `ConfigReader` classes if they are not
            simple types (e.g. int, str, float, Iterable[int|str|float]
            scope: The scope at which to evaluate any child classes. This MUST be passed as scope=globals() if you are
            having any children that are to be parsed.

        Returns:

            An object of cls if ignore_children is True or a tuple with an object of cls and a dictionary with children
            that were not parsed.

        Remarks:

            - This function will return an object of its class after passing the key-value pairs found in the config to
              the init function.

            - Requiring passing scope=globals() to this function is to get around the fact that in python eval() will be
              called with a globals dictionary based on the module in which the function is defined not called. This means
              that in general when eval() is called to create the children, it will not have access to the class
              definitions of these children (except if they happen to be imported in this file). To avoid this problem
              causing an undefined_name exception, the caller must pass her globals() as the scope.

        """
        config = cls.read_config(config=config, section=section)

        if config is None:
            if ignore_children:
                return None
            else:
                return None, {}

        # now we have a dict called config which has our configuration

        myconfig = {}  # parts of the config that can directly be parsed
        children = {}  # parts of the config that need further parsing
        setters = (
            []
        )  # the setters are those configs that have a set_ function for them.

        def _is_simple(x):
            """Tests whether the input can directly be parsed"""
            return (
                x is None
                or isinstance(x, int)
                or isinstance(x, str)
                or isinstance(x, float)
                or (
                    isinstance(x, Iterable)
                    and not isinstance(x, dict)
                    and all(_is_simple(_) for _ in list(x))
                )
            )

        def _set_simple_config(key, v) -> dict[str, Any] | None:
            """Sets a simple value v for key taken into accout its class and the class we are constructing"""
            key_name, class_name = cls._split_key(key)
            _setter = "set_" + key_name
            params = {}
            if hasattr(cls, _setter):
                setters.append((_setter, v))
                return None
            params[key_name] = (
                v
                if class_name is None
                else get_class(class_name=class_name, scope=scope)(v)
            )
            return params

        # read the configs key by key and try to parse anything that is simple enough to parse

        for k, v in config.items():  # type: ignore
            if isinstance(v, dict):
                children[k] = v
            elif isinstance(v, Iterable) and not isinstance(v, str):
                lst = list(v)
                if all(_is_simple(_) for _ in lst):
                    # that is a simple value of the form k:class_name = v. We construct class_name (if it exists) with v
                    # notice that we need to remove class_name when setting the key in myconfig
                    val = _set_simple_config(k, v)
                    if val is not None:
                        myconfig.update(val)
                else:
                    children[k] = v  # type: ignore
            else:
                # that is a simple value of the form k:class_name = v. We construct class_name (if it exists) with v
                val = _set_simple_config(k, v)
                if val is not None:
                    myconfig.update(val)

        # now myconfig has all simply parsed parts and children has all non-parsed parts

        if len(children) > 0 and try_parsing_children:
            if scope is None:
                ValueError(
                    f"scope is None but that is not allowed. You must pass scope=globals() or scope=locals() to "
                    f"from_config. If your classes are defined in the global scope pass globals() and if they "
                    f"are defined in local scope then pass locals(). You can only pass scope=None if you are "
                    f"sure that all of the constructor parameters of the class you are creating are simple "
                    f"values like ints floats and strings."
                )
            parsed_conf, remaining_children, setters = cls._parse_children_config(
                children=children, scope=scope
            )
            myconfig.update(parsed_conf)
            children = remaining_children

        main_object = cls(**myconfig)  # type: ignore

        if try_parsing_children:
            # we will only have setters if we have children
            for setter, value in setters:
                getattr(main_object, setter)(value)

        if ignore_children:
            return main_object
        return main_object, children


class NpEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.integer):
            return int(obj)
        elif isinstance(obj, np.floating):
            return float(obj)
        elif isinstance(obj, np.ndarray):
            return obj.tolist()
        elif isinstance(obj, bytes):
            encoded = base64.b64encode(
                obj
            )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
            return BYTES_START + encoded.decode("ascii")  #
        elif isinstance(obj, Path):
            return PATH_START + str(obj)
        elif not is_jsonable(obj):
            # it may be a type. Always convert types to full names when saving to json
            try:
                obj = TYPE_START + get_full_type_name(obj)
            except:
                return obj
            return obj
        else:
            return super().default(obj)


class NpDecorder(json.JSONDecoder):
    def default(self, obj):
        if isinstance(obj, str):
            if obj.startswith(BYTES_START):
                return base64.b64decode(
                    obj[BYTES_START:]
                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
            elif obj.startswith(TYPE_START):
                return get_class(
                    obj[TYPE_START:]
                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
            elif obj.startswith(PATH_START):
                return Path(
                    obj[PATH_START:]
                )  # b'ZGF0YSB0byBiZSBlbmNvZGVk' (notice the "b")
        return super().default(obj)  #  type: ignore


def dump(
    d: Any,
    file_name: str | os.PathLike | pathlib.Path,
    sort_keys=True,
    compact=False,
) -> None:
    """
    Saves an object depending on the extension of the file given. If the filename given has no extension,
    `DEFAULT_DUMP_EXTENSION` will be used

    Args:
        d: Object to save
        file_name: file name
        sort_keys: If true, the keys will be sorted before saving
        compact: If given, a compact representation will be tried

    Remarks:

        - Supported formats are json, yaml
        - If None is given, the file will be created but will be empty
        - Numpy arrays will be converted to lists before being dumped

    """
    file_name = pathlib.Path(file_name).expanduser().absolute()
    if file_name.suffix == "":
        file_name = pathlib.Path(str(file_name) + "." + DEFAULT_DUMP_EXTENSION)

    if d is None:
        with open(file_name, "w") as f:
            pass
    if file_name.suffix == ".json":
        with open(file_name, "w") as f:
            json.dump(
                d,
                f,
                sort_keys=sort_keys,
                indent=2 if not compact else None,
                cls=NpEncoder,
            )
    elif file_name.suffix in (".yaml", ".yml"):
        with open(file_name, "w") as f:
            yaml.safe_dump(d, f)
    elif file_name.suffix == ".pickle":
        with open(file_name, "wb") as f:
            pickle.dump(d, f)
    elif file_name.suffix == ".csv":
        if not isinstance(d, pd.DataFrame):
            try:
                d = pd.DataFrame(d)
            except Exception as e:
                raise ValueError(f"Failed to convert to a dataframe: {str(e)}")
        d.to_csv(file_name)
    else:
        raise ValueError(f"Unknown extension {file_name.suffix} for {file_name}")


def load(file_name: str | os.PathLike | pathlib.Path) -> Any:
    """
    Loads an object depending on the extension of the file given. If the filename given has no extension,
    `DEFAULT_DUMP_EXTENSION` will be used

    Args:
        file_name: file name

    Remarks:

        - Supported formats are json, yaml
        - If None is given, the file will be created but will be empty

    """
    file_name = pathlib.Path(file_name).expanduser().absolute()
    if file_name.suffix == "":
        file_name = pathlib.Path(str(file_name) + "." + DEFAULT_DUMP_EXTENSION)
    d = {}
    if not file_name.exists() or os.stat(file_name).st_size < 2:
        return d

    if file_name.suffix == ".json":
        with open(file_name) as f:
            d = json.load(f)
    elif file_name.suffix == ".yaml":
        with open(file_name) as f:
            yaml.safe_load(f)
    elif file_name.suffix == ".pickle":
        with open(file_name, "rb") as f:
            d = pickle.load(f)
    elif file_name.suffix == ".csv":
        d = pd.read_csv(file_name).to_dict()  # type: ignore
    else:
        raise ValueError(f"Unknown extension {file_name.suffix} for {file_name}")
    return d


def add_records(
    file_name: str | os.PathLike,
    data: Any,
    col_names: list[str] | None = None,
    raise_exceptions=False,
) -> None:
    """
    Adds records to a csv file

    Args:

        file_name: file name
        data: data to use for creating the record
        col_names: Names in the data.
        raise_exceptions: If given, exceptions  are raised on failure

    Returns:

        None

    Remarks:

        - If col_names are not given, the function will try to normalize the input data if it
          was a dict or a list of dicts

    """
    if col_names is None and (
        isinstance(data, dict)
        or (isinstance(data, list) and len(data) > 0 and isinstance(data[0], dict))
    ):
        data = pd.json_normalize(data)
    else:
        data = pd.DataFrame(data=data, columns=col_names)
    if len(data) < 1:
        return
    file_name = pathlib.Path(file_name)
    file_name.parent.mkdir(parents=True, exist_ok=True)
    new_file = True
    mode = "a"
    # if file_name.exists():
    if is_nonzero_file(file_name):
        new_file = False
        with open(file_name) as f:
            header = f.readline().strip().strip("\n")
        cols = header.split(",")
        for col in cols:
            if len(col) > 0 and col not in data.columns:
                data[col] = None
        if {_ for _ in data.columns} == set(cols):
            data = data.loc[:, cols]
        else:
            try:
                old_data = pd.read_csv(file_name, index_col=None)
                data = pd.concat((old_data, data), axis=0, ignore_index=True)  # type: ignore
            except Exception as e:
                if raise_exceptions:
                    raise e
                warnings.warn(
                    f"Failed to read data from file {str(file_name)} will override it\n{e}",
                    warnings.NegmasIOWarning,
                )

            mode = "w"
            new_file = True

    data.to_csv(str(file_name), index=False, index_label="", mode=mode, header=new_file)



"""
Defines a world for running negotiations directly
"""
from __future__ import annotations

import logging
from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Callable, Collection, Iterable

from negmas import MechanismState, NegotiatorMechanismInterface
from negmas.helpers import get_class, get_full_type_name, instantiate
from negmas.negotiators import Negotiator
from negmas.outcomes import Issue
from negmas.preferences import Preferences
from negmas.serialization import deserialize, serialize, to_flat_dict
from negmas.situated import (
    Action,
    Agent,
    AgentWorldInterface,
    Breach,
    Contract,
    NoContractExecutionMixin,
    RenegotiationRequest,
    World,
)

__all__ = [
    "NegWorld",
    "NegAgent",
    "NegDomain",
]


class NegAgent(Agent):
    """Wraps a negotiator for evaluaton"""

    def __init__(
        self,
        *args,
        negotiator_type: str | type[Negotiator],
        negotiator_params: dict[str, Any] | None = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._negotiator_params = negotiator_params if negotiator_params else dict()
        self._negotiator_type = get_class(negotiator_type)

    @property
    def short_type_name(self):
        """Returns a short name of the type of this entity"""
        return self._negotiator_type.__name__.replace("Negotiator", "").replace(
            "Agent", ""
        )

    @property
    def type_name(self):
        """Returns a short name of the type of this entity"""
        return get_full_type_name(self._negotiator_type)

    @classmethod
    def _type_name(cls):
        return cls.__module__ + "." + cls.__name__

    def make_negotiator(self, preferences: Preferences | None = None):
        """Makes a negotiator of the appropriate type passing it an optional ufun"""
        return instantiate(
            self._negotiator_type, preferences=preferences, **self._negotiator_params
        )

    def _respond_to_negotiation_request(
        self,
        initiator: str,
        partners: list[str],
        issues: list[Issue],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        role: str | None,
        req_id: str | None,
    ) -> Negotiator | None:
        """Responds to any negotiation request by creating a negotiator"""
        return self.make_negotiator(self.awi.get_preferences(partners.index(self.id)))

    def step(self):
        """Called by the simulator at every simulation step"""

    def init(self):
        """Called to initialize the agent **after** the world is initialized. the AWI is accessible at this point."""

    def on_neg_request_rejected(self, req_id: str, by: list[str] | None):
        """Called when a requested negotiation is rejected"""

    def on_neg_request_accepted(
        self, req_id: str, mechanism: NegotiatorMechanismInterface
    ):
        """Called when a requested negotiation is accepted"""

    def on_negotiation_failure(
        self,
        partners: list[str],
        annotation: dict[str, Any],
        mechanism: NegotiatorMechanismInterface,
        state: MechanismState,
    ) -> None:
        """Called whenever a negotiation ends without agreement"""

    def on_negotiation_success(
        self, contract: Contract, mechanism: NegotiatorMechanismInterface
    ) -> None:
        """Called whenever a negotiation ends with agreement"""

    def on_contract_signed(self, contract: Contract) -> None:
        """Called whenever a contract is signed by all partners"""

    def on_contract_cancelled(self, contract: Contract, rejectors: list[str]) -> None:
        """Called whenever at least a partner did not sign the contract"""

    def respond_to_renegotiation_request(
        self, contract: Contract, breaches: list[Breach], agenda: RenegotiationRequest
    ) -> Negotiator | None:
        """Called to respond to a renegotiation request"""

    def on_contract_executed(self, contract: Contract) -> None:
        """
        Called after successful contract execution for which the agent is one of the partners.
        """

    def on_contract_breached(
        self, contract: Contract, breaches: list[Breach], resolution: Contract | None
    ) -> None:
        """
        Called after complete processing of a contract that involved a breach.
        """

    def set_renegotiation_agenda(
        self, contract: Contract, breaches: list[Breach]
    ) -> RenegotiationRequest | None:
        """
        Received by partners in ascending order of their total breach levels in order to set the
        renegotiation agenda when contract execution fails
        """


def _unwrap_negotiators(types, params):
    """
    Removes the agent wrapping the negotiator for each type
    """
    types = list(types)
    old_types = [_ if issubclass(_, NegAgent) else NegAgent for _ in types]
    n = len(types)

    if params is None:
        params = [dict() for _ in range(n)]

    params = [dict() if _ is None else _ for _ in params]

    for i, (t, p) in enumerate(zip(types, params)):
        t = types[i] = get_class(t)
        if issubclass(t, Negotiator):
            continue
        params[i] = p.get("negotiator_params", dict())
        types[i] = p["negotiator_type"]

    return types, params, old_types


def _wrap_in_agents(types, params, agent_types):
    """
    wraps each negotiator in `types` with an agent from `agent_types`
    """
    types = list(types)
    n = len(types)

    if params is None:
        params = [dict() for _ in range(n)]

    if isinstance(agent_types, str):
        agent_types = get_class(agent_types)

    if not isinstance(agent_types, Iterable):
        agent_types = [agent_types] * n

    params = [dict() if _ is None else _ for _ in params]

    for i, (t, p, w) in enumerate(zip(types, params, agent_types)):
        t = types[i] = get_class(t)
        if issubclass(t, NegAgent):
            continue
        name = p.get("name", None)
        params[i] = dict(
            negotiator_type=t, negotiator_params={k: v for k, v in p.items()}
        )
        params[i]["name"] = name
        types[i] = w
    return types, params


@dataclass
class NegDomain:
    """
    A representation of a negotiation domain in which a negotiator can be evaluated
    """

    name: str
    """Domain name"""
    issues: tuple[Issue, ...]
    """The issue space as a list of issues"""
    ufuns: tuple[Preferences, ...]
    """The utility functions used by all negotiators in the domain"""
    partner_types: tuple[str | Negotiator, ...]
    """The types of all partners (other than the agent being evaluated). Its length must be one less than `ufuns`"""
    index: int = 0
    """The index of the negotiator being evaluated in the list of negotiators passed to the mechanism"""
    partner_params: tuple[dict[str, Any] | None, ...] | None = None
    """Any paramters used to construct partners (must be the same length as `partner_types`)"""
    roles: tuple[str, ...] | None = None
    """Roles of all negotiators (includng the negotiator being evaluated) in order"""
    annotation: dict[str, Any] | None = None
    """Any extra annotation to add to the mechanism."""

    def to_dict(self):
        return dict(
            name=self.name,
            issues=[i.to_dict() for i in self.issues],
            ufuns=[serialize(u) for u in self.ufuns],
            partner_types=[get_full_type_name(_) for _ in self.partner_types],  # type: ignore
            index=self.index,
            partner_params=serialize(self.partner_params),
            roles=self.roles,
            annotation=serialize(self.annotation),
        )

    @classmethod
    def from_dict(cls, d):
        return cls(
            name=d["name"],
            issues=tuple(Issue.from_dict(_) for _ in d["issues"]),
            ufuns=tuple(deserialize(u) for u in d["ufuns"]),  # type: ignore The type should be correct in the dict
            partner_types=tuple(get_class(_) for _ in d["partner_types"]),
            index=d["index"],
            partner_params=deserialize(d["partner_params"]),  # type: ignore The type should be correct in the dict
            roles=d["roles"],
            annotation=deserialize(d["annotation"]),  # type: ignore The type should be correct in the dict
        )


class _NegPartner(NegAgent):
    """A `NegAgent` representing a partner that is not being evaluated"""


class _NegAWI(AgentWorldInterface):
    """The AWI for the `NegWorld`"""

    def get_preferences(self, uid: int):
        """Get the agent's ufun"""
        return self._world._domain.ufuns[uid]


class NegWorld(NoContractExecutionMixin, World):
    """
    A world that runs a list of negotiators in a given domain to evaluate them

    Args:
        domain: The `NegDomain` specifying all information about the situation
                in which negotiators are to be evaluated including the partners.
        types: The negotiator types to be evaluated
        params: Any parameters needed to create negotiators
        agent_names_reveal_type: if given the agent name for each negotiator will
                                 simply be the negotiator's type
        compact: If given the system will strive to save memory and minimize logging
        no_logs: disables all logging
        kwargs: Any extra arguments to be passed to the `World` constructor
    """

    def __init__(
        self,
        *args,
        domain: NegDomain,
        types: list[Negotiator | NegAgent],
        params: list[dict[str, Any] | None] | None = None,
        agent_names_reveal_type: bool = True,
        compact: bool = False,
        no_logs: bool = False,
        normalize_scores: bool = False,
        **kwargs,
    ):
        kwargs["log_to_file"] = not no_logs
        if compact:
            kwargs["event_file_name"] = None
            kwargs["event_types"] = []
            kwargs["log_screen_level"] = logging.CRITICAL
            kwargs["log_file_level"] = logging.ERROR
            kwargs["log_negotiations"] = False
            kwargs["log_ufuns"] = False
            # kwargs["save_mechanism_state_in_contract"] = False
            kwargs["save_cancelled_contracts"] = False
            kwargs["save_resolved_breaches"] = False
            kwargs["save_negotiations"] = True
        else:
            kwargs["save_negotiations"] = True

        self.compact = compact
        super().__init__(*args, awi_type="negmas.situated.neg._NegAWI", **kwargs)

        if self.info is None:
            self.info = {}
        self.info["domain"] = serialize(domain)
        self.info["n_steps"] = self.n_steps
        self.info["n_negotiators"] = len(domain.ufuns)

        if domain.annotation is None:
            domain.annotation = dict()
        self._domain = domain
        self._received_utility: dict[str, list[float]] = defaultdict(list)
        self._partner_utility: dict[str, list[float]] = defaultdict(list)
        self._success: dict[str, bool] = defaultdict(bool)
        self._received_advantage: dict[str, list[float]] = defaultdict(list)
        self._n_agreements_per_cometitor: dict[str, list[int]] = defaultdict(list)
        self._partner_advantage: dict[str, list[float]] = defaultdict(list)
        self._competitors: dict[str, NegAgent] = dict()
        self._partners: dict[str, NegAgent] = dict()
        self._n_negs_per_copmetitor = defaultdict(int)
        self._normalize_scores = normalize_scores
        self._preferences_ranges = None
        # if not all(isinstance(_, HasRange) for _ in domain.ufuns):
        #     raise ValueError(f"Not all ufuns has a range!!!")
        if self._normalize_scores:
            self._preferences_ranges = [
                u.minmax() for u in domain.ufuns  # type: ignore We already check just above
            ]
        partner_types = domain.partner_types
        partner_params = domain.partner_params

        if partner_params is None:
            partner_params = [dict() for _ in partner_types]

        for i, p in enumerate(partner_params):
            if p is None:
                partner_params[i] = dict()

        types = list(types)
        len(types)

        types, params, wrappers = _unwrap_negotiators(types, params)

        for i, t in enumerate(types):
            types[i] = get_full_type_name(t)

        self.agent_unique_types = [
            f"{t}:{hash(str(p)) if p else ''}" if len(p) > 0 else t
            for t, p in zip(types, params)
        ]

        def add_agents(types, params, wrappers, target):
            types, params = _wrap_in_agents(types, params, wrappers)
            for t, p in zip(types, params):
                agent = instantiate(t, **p)
                if agent_names_reveal_type:
                    agent.name = agent.short_type_name
                    agent.id = agent.short_type_name
                target[agent.id] = agent
                self.join(agent)

        add_agents(types, params, wrappers, self._competitors)

        add_agents(partner_types, partner_params, _NegPartner, self._partners)

    def simulation_step(self, stage):
        def unormalize(u, indx):
            if self._normalize_scores:
                mn, mx = self._preferences_ranges[indx]
                if mx > mn:
                    u = (u - mn) / (mx - mn)
                else:
                    u = u / mx
            return u

        def calcu(ufun, indx, agreement):
            u = ufun(agreement)
            return unormalize(u, indx)

        for aid, agent in self._competitors.items():
            partners = list(self._partners.keys())
            partners.insert(self._domain.index, aid)
            _, mechanism = self.run_negotiation(
                caller=agent,
                issues=self._domain.issues,
                partners=partners,
                roles=self._domain.roles,
                annotation=self._domain.annotation,
                negotiator=None,
            )
            self._n_negs_per_copmetitor[aid] += 1
            agreement = mechanism.state.agreement if mechanism else None
            self._success[aid] = mechanism is not None
            agent = self.agents[aid]
            ufun = self._domain.ufuns[self._domain.index]
            u = float(
                calcu(ufun, self._domain.index, agreement)
                if agreement
                else unormalize(ufun.reserved_value, self._domain.index)
            )
            r = unormalize(float(ufun.reserved_value), self._domain.index)
            self._received_utility[aid].append(u)
            self._n_agreements_per_cometitor[aid].append(int(mechanism is not None))
            self._received_advantage[aid].append(u - r)
            pufuns = [
                (partners.index(pid), p.awi.get_preferences(partners.index(pid)))
                for pid, p in self._partners.items()
            ]
            pu = sum(unormalize(float(_(agreement)), i) for i, _ in pufuns)
            pa = sum(
                unormalize(float(_(agreement)), i)
                - unormalize(float(_.reserved_value), i)
                for i, _ in pufuns
            )
            self._partner_utility[aid].append(pu)
            self._partner_advantage[aid].append(pa)
            partner_names = [self.agents[_].name for _ in partners]
            self.loginfo(f"{agent.name} : {partner_names} -> {agreement}")

    def received_utility(self, aid: str):
        return sum(
            self._received_utility.get(aid, [0])
        ) / self._n_negs_per_copmetitor.get(aid, 0)

    def agreement_rate(self, aid: str):
        return sum(
            self._n_agreements_per_cometitor.get(aid, [0])
        ) / self._n_negs_per_agent.get(aid, 0)

    def partner_utility(self, aid: str):
        return sum(
            self._partner_utility.get(aid, [0])
        ) / self._n_negs_per_copmetitor.get(aid, 0)

    def received_advantage(self, aid: str):
        return sum(
            self._received_advantage.get(aid, [0])
        ) / self._n_negs_per_copmetitor.get(aid, 0)

    def partner_advantage(self, aid: str):
        return sum(
            self._partner_advantage.get(aid, [0])
        ) / self._n_negs_per_copmetitor.get(aid, 0)

    @property
    def competitors(self):
        return self._competitors

    @property
    def partners(self):
        return self._partners

    def post_step_stats(self):
        for aid in self._competitors.keys():
            self._stats[f"has_agreement_{aid}"].append(self._success[aid])
            self._stats[f"received_utility_{aid}"].append(
                self._received_utility[aid][-1]
            )
            self._stats[f"partner_utility_{aid}"].append(self._partner_utility[aid][-1])
            self._stats[f"received_advantage_{aid}"].append(
                self._received_advantage[aid][-1]
            )
            self._stats[f"agrement_rate_{aid}"].append(
                self._n_agreements_per_cometitor[aid][-1]
            )
            self._stats[f"partner_advantage_{aid}"].append(
                self._partner_advantage[aid][-1]
            )

    def pre_step_stats(self):
        pass

    def order_contracts_for_execution(
        self, contracts: Collection[Contract]
    ) -> Collection[Contract]:
        return contracts

    def contract_record(self, contract: Contract) -> dict[str, Any]:
        return to_flat_dict(contract, deep=True)

    def breach_record(self, breach: Breach) -> dict[str, Any]:
        return to_flat_dict(breach, deep=True)

    def contract_size(self, contract: Contract) -> float:
        n = len(self._competitors)
        return 1.0 / (n if n else 1)

    def delete_executed_contracts(self) -> None:
        pass

    def execute_action(
        self, action: Action, agent: Agent, callback: Callable | None = None
    ) -> bool:
        """Executes the given action by the given agent"""
        ...

    def get_private_state(self, agent: Agent) -> dict:
        """Reads the private state of the given agent"""
        ...

    def executable_contracts(self) -> Collection[Contract]:
        return []

    def start_contract_execution(self, contract: Contract) -> set[Breach]:
        return set()

    def complete_contract_execution(
        self, contract: Contract, breaches: list[Breach], resolution: Contract
    ) -> None:
        pass


if __name__ == "__main__":
    from negmas.genius import genius_bridge_is_running
    from negmas.genius.gnegotiators import Atlas3, NiceTitForTat
    from negmas.outcomes import Issue
    from negmas.preferences import LinearAdditiveUtilityFunction as U
    from negmas.sao import AspirationNegotiator, NaiveTitForTatNegotiator
    from negmas.situated import save_stats

    issues = [Issue(10, "quantity"), Issue(5, "price")]
    competitors = [AspirationNegotiator, NaiveTitForTatNegotiator]
    if genius_bridge_is_running():
        competitors += [Atlas3, NiceTitForTat]

    domain = NegDomain(
        name="d0",
        issues=issues,
        ufuns=[
            U.random(issues, reserved_value=(0.0, 0.2), normalized=True),
            U.random(issues, reserved_value=(0.0, 0.2), normalized=True),
        ],
        partner_types=[AspirationNegotiator],
        index=0,
    )
    world = NegWorld(
        domain=domain,
        types=competitors,
        n_steps=2,
        neg_n_steps=10,
        neg_time_limit=None,
    )
    world.run()
    print("World ran", flush=True)
    for aid in world.agents.keys():
        print(world.received_utility(aid))
    save_stats(world, world.log_folder)



from __future__ import annotations

import copy
import time
from abc import abstractmethod
from typing import Callable

import numpy as np

from ..common import MechanismState, NegotiatorMechanismInterface, Value
from ..helpers.prob import ScipyDistribution
from ..models.acceptance import AdaptiveDiscreteAcceptanceModel, DiscreteAcceptanceModel
from ..outcomes import Outcome
from ..preferences import IPUtilityFunction, MappingUtilityFunction, Preferences
from ..sao import AspirationNegotiator, ResponseType, SAONegotiator
from .common import _locs, _uppers
from .expectors import Expector, MeanExpector
from .strategy import EStrategy
from .user import User

__all__ = ["BaseElicitor"]


class BaseElicitor(SAONegotiator):
    def accuracy_limit(self, cost: float) -> float:
        """The accuracy limit given the cost and `epsilon`."""
        return 0.5 * max(self.epsilon, cost)

    def __init__(
        self,
        user: User,
        *,
        strategy: EStrategy | None = None,
        base_negotiator: SAONegotiator = AspirationNegotiator(),
        opponent_model_factory: None
        | (
            Callable[[NegotiatorMechanismInterface], DiscreteAcceptanceModel]
        ) = lambda x: AdaptiveDiscreteAcceptanceModel.from_negotiation(nmi=x),
        expector_factory: Expector | Callable[[], Expector] = MeanExpector,
        single_elicitation_per_round=False,
        continue_eliciting_past_reserved_val=False,
        epsilon=0.001,
        true_utility_on_zero_cost=False,
    ) -> None:
        """
        ABC for all elicitation algorithms.

        Args:
            user: A `User` object that
            strategy: An elicitation strategy that determines the order of deep
                      elicitation queries.
            base_negotiator: A negotiator that is used to propose and rspond to
                      proposals based on the current state of the utility function.
            opponent_model_factory: A callable that can be used to create an opponent
                                    model.
            expector_factory: A callable that can be used to construct an `Expector`
                              object responsible of reducing a probabilistic utility
                              value into a real number to be used by the negotiator.
            single_elicitation_per_round: Forces a single elicitation opportunity per
                                          negotiation round. If the elicitor uses deep
                                          elicitation this will correspond to multiple
                                          calls to the `strategy`.
            continue_eliciting_past_reserved_val: If `True`, elicitation continues even
                                                  if the expector returns a value under
                                                  the reserved value of the negotiator.
            epsilon: A small number to stop elicitation when the uncertainty in the ufun
                     is under.
            true_utility_on_zero_cost: If `True`, zero cost will force the final elicited
                                       value of any outcome to exactly match the utility
                                       function. If `False`, the final utility value after
                                       elicitation may be within `epsilon` from the true
                                       value.
        """
        super().__init__()
        self.add_capabilities(
            {
                "propose": True,
                "respond": True,
                "propose-with-value": False,
                "max-proposals": None,  # indicates infinity
            }
        )
        self.strategy = strategy
        self.opponent_model_factory = opponent_model_factory
        self.expector_factory = expector_factory
        self.single_elicitation = single_elicitation_per_round
        self.continue_eliciting_past_reserved_val = continue_eliciting_past_reserved_val
        self.epsilon = epsilon
        self.true_utility_on_zero_cost = true_utility_on_zero_cost
        self.elicitation_history = []
        self.opponent_model = None
        self._elicitation_time = None
        self.asking_time = 0.0
        self.offerable_outcomes = (
            []
        )  # will contain outcomes with known or at least elicited utilities
        self.indices = None
        self.initial_utility_priors = None
        self.user = user
        self.acc_limit = self.accuracy_limit(self.user.cost_of_asking())
        self.base_negotiator = base_negotiator
        self.expect = None
        if strategy is not None:
            strategy.resolution = max(self.acc_limit, strategy.resolution)

    def init_elicitation(
        self,
        preferences: None
        | (IPUtilityFunction | ScipyDistribution | list[ScipyDistribution]),
        **kwargs,
    ) -> None:
        """
        Called once to initialize the elicitation process

        Args:
            preferences: The probabilistic utility function
            **kwargs:

        Remarks:
            - If no `ufun` is given one will be created with 0-1 uniform distributions and
              zero reserved value.
            - If a single `Distribution` is given as `ufun`, it is repeated for all
              outcomes (and the reserved value is set to zero).
            - If a list of `Distribution` s is given, it must have the same length as
              the list of outcomes of this negotiation and is used to set the `ufun`.
            - The opponent model

        """
        nmi = self._nmi
        if nmi is None:
            raise ValueError(f"Unkown NMI")
        self.elicitation_history = []
        outcomes = list(nmi.discrete_outcomes())
        self.indices = dict(zip(outcomes, range(len(outcomes))))
        self.offerable_outcomes = []
        self._elicitation_time = 0.0
        if self.opponent_model_factory is None:
            self.opponent_model = None
        else:
            self.opponent_model = self.opponent_model_factory(nmi)
            self.base_negotiator.opponent_model = self.opponent_model
        if preferences is None:
            dists = [
                ScipyDistribution(type="uniform", loc=0.0, scale=1.0) for _ in outcomes
            ]
            preferences = IPUtilityFunction(
                outcomes=outcomes, distributions=dists, reserved_value=0.0
            )
        elif isinstance(preferences, ScipyDistribution):
            preferences = [copy.copy(preferences) for _ in outcomes]
            preferences = IPUtilityFunction(
                outcomes=outcomes, distributions=preferences, reserved_value=0.0
            )
        elif (
            isinstance(preferences, list)
            and len(preferences) > 0
            and isinstance(preferences[0], ScipyDistribution)
        ):
            preferences = IPUtilityFunction(
                outcomes=outcomes, distributions=preferences, reserved_value=0.0
            )
        self.set_preferences(preferences)
        self.initial_utility_priors = copy.copy(preferences)

    def join(
        self,
        nmi: NegotiatorMechanismInterface,
        state: MechanismState,
        *,
        preferences: Preferences | None = None,
        role: str = "negotiator",
        **kwargs,
    ) -> bool:
        """
        Called to join a negotiation.

        Remarks:
            - uses the base_negotiator to join the negotiation.
            - creates a `MappingUtilityFunction` that maps every outcome to the
              result of the expector applied to the corresponding utility value.
            - The reserved value of the created ufun is set to -inf
        """
        if preferences is None:
            preferences = IPUtilityFunction(outcomes=nmi.outcomes, reserved_value=0.0)
        if not super().join(nmi=nmi, state=state, preferences=preferences, role=role):
            return False
        self.expect = self.expector_factory(self._nmi)
        self.init_elicitation(preferences=preferences, **kwargs)
        self.base_negotiator.join(
            nmi,
            state,
            preferences=MappingUtilityFunction(
                mapping=lambda x: self.expect(self.preferences(x), state=state),
                reserved_value=float("-inf"),
            ),
        )
        return True

    def on_negotiation_start(self, state: MechanismState):
        """Called when the negotiation starts. Just passes the call to
        base_negotiator."""
        self.base_negotiator.on_negotiation_start(state=state)

    def utility_distributions(self) -> list[ScipyDistribution]:
        """
        Returns a `Distribution` for every outcome
        """
        if self.preferences is None:
            return [None] * len(self._nmi.outcomes)
        if self.preferences.base_type == "ip":
            return list(self.preferences.distributions.values())
        else:
            return [self.preferences(o) for o in self._nmi.outcomes]

    def user_preferences(self, outcome: Outcome | None) -> float:
        """
        Finds the total utility obtained by the user for this outcome after
        discounting elicitation cost.

        Args:
            outcome: The outcome to find the user utility for. If None, it
                     returns the reserved value.

        Remarks:
            The total elicitation cost is *not* discounted from the reserved
            value when the input is None
        """
        return (
            self.user.ufun(outcome) - self.user.total_cost
            if outcome is not None
            else self.user.ufun(outcome)
        )

    @property
    def elicitation_cost(self) -> float:
        """
        The total elicitation cost.
        """
        return self.user.total_cost

    @property
    def elicitation_time(self) -> float:
        """The total elicitation time in seconds."""
        return self._elicitation_time

    def maximum_attainable_utility(self) -> float:
        """
        Maximum utility that could even in principle be attained which
        simply means the utility value of the outcome with maximum utility.
        """
        return max(_uppers(self.utility_distributions()))

    def minimum_guaranteed_utility(self):
        """
        Minimum utility that could even in principle be attained which
        simply means the utility value of the outcome with minimum utility.
        """
        return min(_locs(self.utility_distributions()))

    def on_opponent_model_updated(
        self, outcomes: list[Outcome], old: list[float], new: list[float]
    ) -> None:
        """
        Called whenever an opponents model is updated.

        Args:
            outcomes: A list of outcomes for which the acceptance probability are changed
            old: The old acceptance probability
            new: The new acceptance probability
        """

    def on_partner_proposal(
        self, state: MechanismState, partner_id: str, offer: Outcome
    ):
        """
        Called when one of the partners propose (only if enable_callbacks is set
        in the `SAOMechanism`).

        Args:
            state: mechanism state
            partner_id: the partner who proposed
            offer: The offer from the partner

        Remarks:
            - Used to update the opponent model by calling `update_offered` then
              `on_opponent_model_updated`.
        """
        self.base_negotiator.on_partner_proposal(
            partner_id=partner_id, offer=offer, state=state
        )
        old_prob = self.opponent_model.probability_of_acceptance(offer)
        self.opponent_model.update_offered(offer)
        new_prob = self.opponent_model.probability_of_acceptance(offer)
        self.on_opponent_model_updated([offer], old=[old_prob], new=[new_prob])

    def on_partner_response(
        self,
        state: MechanismState,
        partner_id: str,
        outcome: Outcome,
        response: ResponseType,
    ):
        """
        Called when one of the partners respond  (only if enable_callbacks is set
        in the `SAOMechanism`).

        Args:
            state: mechanism state
            partner_id: the partner who offered
            outcome: The outcome responded to
            response: The partner response including both the response and outcome

        Remarks:
            - Used to update the opponent model by calling `update_rejected` or
              `update_accepted1 then `on_opponent_model_updated`.
        """
        self.base_negotiator.on_partner_response(
            state=state, partner_id=partner_id, outcome=outcome, response=response
        )
        if response == ResponseType.REJECT_OFFER:
            old_probs = [self.opponent_model.probability_of_acceptance(outcome)]
            self.opponent_model.update_rejected(outcome)
            new_probs = [self.opponent_model.probability_of_acceptance(outcome)]
            self.on_opponent_model_updated([outcome], old=old_probs, new=new_probs)
        elif response == ResponseType.ACCEPT_OFFER:
            old_probs = [self.opponent_model.probability_of_acceptance(outcome)]
            self.opponent_model.update_accepted(outcome)
            new_probs = [self.opponent_model.probability_of_acceptance(outcome)]
            self.on_opponent_model_updated([outcome], old=old_probs, new=new_probs)

    def before_eliciting(self) -> None:
        """Called by apply just before continuously calling elicit_single"""

    @abstractmethod
    def elicit_single(self, state: MechanismState) -> None:
        """Does a single elicitation act

        Args:
            state: mechanism state
        """
        raise NotImplementedError()

    def elicit(self, state: MechanismState) -> None:
        """
        Called to do utility elicitation whenever needed.

        Args:
            state: mechanism state

        Remarks:
            - Keeps track of elicitation time and asking time.
            - If the maximum attainable utility minus elicitation cost is less than the
              reserved value, no elicitation will take place because we will end this
              negotiation anyway. Note that the maximum attainable utility can **never**
              go up.
            - Calls `before_eliciting` once to initialize the process then calls
              `elicit_single` which does the actual elicitation. This is done only once
              if `single_elicitation` is set, otherwise it is repeated until one of the
              following conditiosn is met:

                - `elicit_single` returns False
                - The maximum attainable utility (minus elicitation cost) is less than
                  the reserved value.
        """
        if (
            self.maximum_attainable_utility() - self.elicitation_cost
            <= self.reserved_value
        ):
            return
        start = time.perf_counter()
        self.before_eliciting()
        if self.single_elicitation:
            self.elicit_single(state=state)
        else:
            while self.elicit_single(state=state):
                if (
                    self.maximum_attainable_utility() - self.elicitation_cost
                    <= self.reserved_value
                    or state.relative_time >= 1
                ):
                    break
        elapsed = time.perf_counter() - start
        self._elicitation_time += elapsed
        self.asking_time += elapsed

    def utility_on_rejection(self, outcome: Outcome, state: MechanismState) -> Value:
        """Estimated utility if this outcome rejected at this state.

        Args:
            outcome: The outcome tested
            state: The mechanism state

        Remarks:
            - MUST be implemented by any Elicitor.
        """
        raise NotImplementedError(
            f"Must override utility_on_rejection in {self.__class__.__name__}"
        )

    def offering_utility(self, outcome, state) -> Value:
        """
        returns expected utlity of offering `outcome` in `state`

        Args:
            outcome: The outcome
            state: The state

        Returns:
            A utility value

        Remarks:
            - returns $u(o) p(o) + ru(o) (1-p(o))$ where $p$ is the opponent model,
              $u$ is the utility function, and $r$ is the utility in case of rejections.
            - `state` is needed when calculating $r(o)$ by calling `utility_on_rejection`.
            - Note that if $u$ or $r$ return a `Distribution`, this method will return
              a `Distribution` not a real number.
        """
        if self.opponent_model is None:
            return self.preferences(outcome)
        u = self.preferences(outcome)
        p = self.opponent_model.probability_of_acceptance(outcome)
        return p * u + (1 - p) * self.utility_on_rejection(outcome, state=state)

    def best_offer(self, state: MechanismState) -> tuple[Outcome | None, float]:
        """The outcome with maximum expected utility given the expector and its utility

        Args:
            state: The mechanism state

        Returns:
            A tuple containing the best outcome (or None) and its expected utility using the
            expector (or reserved value)

        Remarks:
            - if there are no offerable outcomes, elicitation is done and if still there are no
              offerable outcomes, the reserved value is returned (with None as outcome)
            - Only offerable outcomes are checked.
            - The best outcome is defined as the one with maximum `expect` applied to
              `offering_utility`.
        """
        if len(self.offerable_outcomes) == 0:
            self.elicit(state=state)
        if len(self.offerable_outcomes) == 0:
            return None, self.reserved_value
        best, best_utility, bsf = None, self.reserved_value, self.reserved_value
        for outcome in self.offerable_outcomes:
            if outcome is None:
                continue
            utilitiy = self.offering_utility(outcome, state=state)
            expected_utility = self.expect(utilitiy, state=state)
            if expected_utility >= bsf:
                best, best_utility, bsf = outcome, utilitiy, expected_utility
        return best, self.expect(best_utility, state=state)

    def respond_(self, state: MechanismState, offer: Outcome) -> ResponseType:
        """
        Called by the mechanism directly (through `counter` ) to respond to offers.

        Args:
            state: mechanism state
            offer: the offer to respond to

        Remarks:
            - Does the following steps:

                1. Finds the the best offer using `best_offer` and uses the base negotiator
                   to respond if that offer was `None`
                2. Looks at `offerable_outcomes` and applies the elicitation strategy (one
                   step) to the outcome if it was not offerable (or if there are no offerable
                   outcomes defined).
                3. Finds the utility of the offer using `utility_function` not taking into accout
                   elicitation cost and uses the base negotiator if that fails (i.e. `utility_function`
                   returns `None`).
                4. Finds the expected utility of the offer using the `expect` () method which calls the
                   expector passed during construction.
                5. If the maximum attainable utility now (judging from the current estimation of
                   the utility value of each outcome taking elicitation cost into account) is less
                   than the reserved value, end the negotiation
                6. If the utility of my best offer (returned from `best_offer`) is less than the offered
                   utility, accept the offer
                7. Otherwise, call bhe base negotiator to respond.

        """
        my_offer, meu = self.best_offer(state=state)
        if my_offer is None:
            return self.base_negotiator.respond_(state=state, offer=offer)
        if (
            self.strategy
            and self.offerable_outcomes is not None
            and offer not in self.offerable_outcomes
        ):
            self.strategy.apply(user=self.user, outcome=offer)
        offered_utility = self.preferences(offer)
        if offered_utility is None:
            return self.base_negotiator.respond_(state=state, offer=offer)
        offered_utility = self.expect(offered_utility, state=state)
        if (
            self.maximum_attainable_utility() - self.user.total_cost
            < self.reserved_value
        ):
            return ResponseType.END_NEGOTIATION
        if meu < offered_utility:
            return ResponseType.ACCEPT_OFFER
        else:
            return self.base_negotiator.respond_(state=state, offer=offer)

    @abstractmethod
    def can_elicit(self) -> bool:
        """Returns whether we can do more elicitation"""
        raise NotImplementedError()

    def propose(self, state: MechanismState) -> Outcome:
        """
        Called to propose an outcome

        Args:
            state: mechanism state

        Remarks:

            - if the negotiator `can_elicit`, it will `elicit`.
            - always then calls the base negotiator to propose.
        """
        if self.can_elicit():
            self.elicit(state=state)
        return self.base_negotiator.propose(state=state)

    def offering_utilities(self, state) -> np.ndarray:
        """
        Calculates the offering utility for all outcomes

        Args:
            state: Calculates the state at which the offering utilities are to be calculated

        Returns:
            An ndarray with the offering utility of every outcome (in order)

        Remarks:
            - This is just a faster version of calling `offering_utility` in a loop.

        """
        us = np.asarray(self.utility_distributions())
        ps = np.asarray(self.opponent_model.acceptance_probabilities())
        return ps * us + (1 - ps) * np.asarray(self.preferences(state=state))

    def utility_on_acceptance(self, outcome: Outcome) -> Value:
        """
        The utility of acceptance which is simply the utility function applied to `outcome`.
        """
        return self.preferences(outcome)

    def utilities_on_rejection(self, state: MechanismState) -> list[Value]:
        """Finds the utility of rejection for all outputs.

        Remarks:
            - By default it calls `utility_on_rejection` repeatedly for all outcomes.
              Override this method if a faster versin can be implemented
        """
        return [
            self.utility_on_rejection(outcome=outcome, state=state)
            for outcome in self._nmi.outcomes
        ]

    def __getattr__(self, item):
        return getattr(self.base_negotiator, item)
        # TODO extend this to take the partner_id as a parameter to handle multiparty negotiation

    def __str__(self):
        return f"{self.name}"



"""
Modeling self's future prospects in the negotiation.
"""
from __future__ import annotations

__all__ = ["FutureUtilityRegressor"]
import numpy as np
from sklearn.gaussian_process import GaussianProcessRegressor


class FutureUtilityRegressor:
    """Represents a regressor for own-utility for the future of the negotiation.

    Remarks:
        - We assume that the negotiation goes from time 0 to 1 (relative_time).
    """

    def __init__(self, regressor_factory=GaussianProcessRegressor, **kwargs):
        self.regressor = regressor_factory(**kwargs)
        self.inverse_regressor = regressor_factory(**kwargs)

    def fit(self, times, utils) -> FutureUtilityRegressor:
        times, utils = np.array(times), np.array(utils)
        times = times.flatten().reshape((len(times), 1))
        utils = utils.flatten().reshape((len(utils), 1))
        self.regressor.fit(times, utils)
        self.inverse_regressor.fit(utils, times)
        return self.inverse_regressor

    def predict_utility(self, times) -> np.ndarray:
        times = np.array(times)
        times = times.flatten().reshape((len(times), 1))
        return self.regressor.predict(times).flatten()

    def predict_utility_prob(self, times, return_cov=False) -> np.ndarray:
        times = times.flatten().reshape((1, len(times)))
        return self.regressor.predict(times, return_std=True, return_cov=return_cov)

    def predict_time(self, utils) -> np.ndarray:
        utils = np.array(utils)
        utils = utils.flatten().reshape((1, len(utils)))
        return self.inverse_regressor.predict(utils).flatten()

    def predict_time_prob(self, utils, return_cov=False) -> np.ndarray:
        utils = np.array(utils)
        utils = utils.flatten().reshape((1, len(utils)))
        return self.inverse_regressor.predict(
            utils, return_std=True, return_cov=return_cov
        )



from __future__ import annotations

from typing import (
    TYPE_CHECKING,
    Collection,
    Container,
    Iterable,
    Protocol,
    runtime_checkable,
)

if TYPE_CHECKING:
    from .base_issue import DiscreteIssue, Issue
    from .common import Outcome
    from .outcome_space import CartesianOutcomeSpace

__all__ = [
    "OutcomeSpace",
    "DiscreteOutcomeSpace",
    "IndependentIssuesOS",
    "IndependentDiscreteIssuesOS",
]


@runtime_checkable
class OutcomeSpace(Container, Protocol):
    """
    The base protocol for all outcome spaces.
    """

    def is_valid(self, outcome: Outcome) -> bool:
        """Checks if the given outcome is valid for that outcome space"""
        ...

    def are_types_ok(self, outcome: Outcome) -> bool:
        """Checks if the type of each value in the outcome is correct for the given issue"""
        ...

    def ensure_correct_types(self, outcome: Outcome) -> Outcome:
        """Returns an outcome that is guaratneed to have correct types or raises an exception"""
        ...

    @property
    def cardinality(self) -> int | float:
        """The space cardinality = the number of outcomes"""
        ...

    def is_numeric(self) -> bool:
        """Checks whether all values in all outcomes are numeric"""
        ...

    def is_integer(self) -> bool:
        """Checks whether all values in all outcomes are integers"""
        ...

    def is_float(self) -> bool:
        """Checks whether all values in all outcomes are real"""
        ...

    def to_discrete(
        self, levels: int | float = 5, max_cardinality: int | float = float("inf")
    ) -> DiscreteOutcomeSpace:
        """
        Returns a **stable** finite outcome space. If the outcome-space is already finite. It shoud return itself.

        Args:
            levels: The levels of discretization of any continuous dimension (or subdimension)
            max_cardintlity: The maximum cardinality allowed for the resulting outcomespace (if the original OS was infinite).
                             This limitation is **NOT** applied for outcome spaces that are alredy discretized. See `limit_cardinality()`
                             for a method to limit the cardinality of an already discrete space

        If called again, it should return the same discrete outcome space every time.
        """
        ...

    def sample(
        self,
        n_outcomes: int,
        with_replacement: bool = False,
        fail_if_not_enough=False,
    ) -> Iterable[Outcome]:
        """Samples up to n_outcomes with or without replacement"""
        ...

    def to_largest_discrete(
        self, levels: int, max_cardinality: int | float = float("inf"), **kwargs
    ) -> DiscreteOutcomeSpace:
        ...

    def cardinality_if_discretized(
        self,
        levels: int,
        max_cardinality: int | float = float("inf"),
    ) -> int:
        """
        Returns the cardinality if discretized the given way.
        """
        ...

    def enumerate_or_sample(
        self,
        levels: int | float = float("inf"),
        max_cardinality: int | float = float("inf"),
    ) -> Iterable[Outcome]:
        """Enumerates all outcomes if possible (i.e. discrete space) or returns `max_cardinality` different outcomes otherwise"""
        ...

    def is_discrete(self) -> bool:
        """Checks whether there are no continua components of the space"""
        ...

    def is_finite(self) -> bool:
        """Checks whether the space is finite"""
        return self.is_discrete()

    def __contains__(self, item: Outcome | OutcomeSpace | Issue) -> bool:
        ...


@runtime_checkable
class DiscreteOutcomeSpace(OutcomeSpace, Collection, Protocol):
    """
    The base protocol for all outcome spaces with a finite number of items.

    This type of outcome-space acts as a standard python `Collection` which
    means that its length can be found using `len()` and it can be iterated over
    to return outcomes.
    """

    @property
    def cardinality(self) -> int:
        """The space cardinality = the number of outcomes"""
        ...

    def enumerate(self) -> Iterable[Outcome]:
        """
        Enumerates the outcome space returning all its outcomes (or up to max_cardinality for infinite ones)
        """
        ...

    def limit_cardinality(
        self,
        max_cardinality: int | float = float("inf"),
        levels: int | float = float("inf"),
    ) -> DiscreteOutcomeSpace:
        """
        Limits the cardinality of the outcome space to the given maximum (or the number of levels for each issue to `levels`)

        Args:
            max_cardinality: The maximum number of outcomes in the resulting space
            levels: The maximum levels allowed per issue (if issues are defined for this outcome space)
        """
        ...

    def to_single_issue(
        self, numeric: bool = False, stringify: bool = True
    ) -> CartesianOutcomeSpace:
        ...

    def sample(
        self,
        n_outcomes: int,
        with_replacement: bool = False,
        fail_if_not_enough=True,
    ) -> Iterable[Outcome]:
        ...

    def is_discrete(self) -> bool:
        """Checks whether there are no continua components of the space"""
        return True

    def to_discrete(self, *args, **kwargs) -> DiscreteOutcomeSpace:
        return self

    def __iter__(self):
        return self.enumerate().__iter__()

    def __len__(self) -> int:
        return self.cardinality


@runtime_checkable
class IndependentIssuesOS(Protocol):
    """
    An Outcome-Space that is constructed from a tuple of `Issue` objects.
    """

    issues: tuple[Issue, ...]


@runtime_checkable
class IndependentDiscreteIssuesOS(Protocol):
    """
    An Outcome-Space that is constructed from a tuple of `DiscreteIssue` objects.
    """

    issues: tuple[DiscreteIssue, ...]
